import sys
import asyncio

import soyutnet
from soyutnet.pt_common import PTRegistry
from soyutnet.token import Token
from soyutnet.place import SpecialPlace, Place
from soyutnet.transition import Transition
from soyutnet.common import GENERIC_ID, GENERIC_LABEL
from soyutnet.observer import ComparativeObserver, Observer


def main(token_count=12):
    token_ids = [GENERIC_ID] * token_count

    async def producer(place):
        try:
            id: id_t = token_ids.pop(0)
            return (GENERIC_LABEL, id)
        except IndexError:
            pass

        return tuple()

    async def consumer(place):
        return

    place_count = 3

    def on_comparison_ends(observer):
        nonlocal place_count
        place_count -= 1
        if place_count == 0:
            soyutnet.terminate()

    o00 = ComparativeObserver(
        expected={2: [-1] + [1] * (len(token_ids) // 2 - 2)},
        on_comparison_ends=on_comparison_ends,
        verbose=False,
    )
    o01 = ComparativeObserver(
        expected={2: [-1] + [1] * (len(token_ids) // 2 - 2)},
        on_comparison_ends=on_comparison_ends,
        verbose=False,
    )
    o1 = ComparativeObserver(
        expected={2: [1] * (len(token_ids) // 2 - 1)},
        on_comparison_ends=on_comparison_ends,
        verbose=False,
    )
    p00 = SpecialPlace("p00", producer=producer, observer=o00)
    p01 = SpecialPlace("p01", producer=producer, observer=o01)
    p1 = SpecialPlace("p1", consumer=consumer, observer=o1)
    t1 = Transition("t1")

    reg = PTRegistry()

    reg.register(p00)
    reg.register(p01)
    reg.register(p1)
    reg.register(t1)

    p00.connect(t1)
    p01.connect(t1)
    t1.connect(p1)

    asyncio.run(soyutnet.main(reg))


if __name__ == "__main__":
    main(int(sys.argv[1]))



import asyncio
import signal
import functools
from typing import Any

from .pt_common import PTRegistry, PTCommon


def _int_handler(
    signame: str, loop: asyncio.AbstractEventLoop, pt_registry: PTRegistry
) -> None:
    print(f"Got signal '{signame}'")

    if signame == "SIGINT" or signame == "SIGTERM":
        print(">>>>>>>>>>>>>>>>>>>>>>>>> <<<<<<<<<<<<<<<<<<<<<<<<<")
        loop.stop()


def _add_int_handlers(pt_registry: PTRegistry) -> None:
    loop = asyncio.get_running_loop()

    for signame in {"SIGINT", "SIGTERM"}:
        loop.add_signal_handler(
            getattr(signal, signame),
            functools.partial(_int_handler, signame, loop, pt_registry),
        )


def _cancel_all_tasks() -> None:
    tasks: set[asyncio.Task[Any]] = asyncio.all_tasks()
    for task in tasks:
        task.cancel()


def terminate() -> None:
    """
    Terminates PT net simulation.
    """
    _cancel_all_tasks()


async def main(pt_registry: PTRegistry, debug_level: str = "ERROR") -> None:
    """
    Main entry point of PT net simulation.

    Runs the tasks assigned to places and transitions registered in ``pt_registry``.

    :param pt_registry: Registry object keeping all places and transitions in the model.
    :param debug_level: Set debug level to ``"ERROR", "DEBUG"`` or ``"INFO"``.
    """
    from . import global_defs

    if debug_level == "INFO":
        global_defs.VERBOSE_ENABLED = True
        debug_level = "DEBUG"
    if debug_level == "DEBUG":
        global_defs.DEBUG_ENABLED = True

    tasks: set[asyncio.Task[PTCommon]] = set()

    _add_int_handlers(pt_registry)
    for loop in pt_registry.get_loops():
        task: asyncio.Task[Any] = asyncio.create_task(loop)
        tasks.add(task)
        task.add_done_callback(tasks.discard)

    await asyncio.gather(*tasks, return_exceptions=False)



from typing import (
    Tuple,
    Never,
    Dict,
)

LOOP_DELAY = 0.0
"""Asyncio tasks main loop delay. It can be picked a positive number for debugging."""
DEBUG_ENABLED = False
"""if set, :py:func:`soyutnet.common.DEBUG` will print."""
VERBOSE_ENABLED = False
"""if set, :py:func:`soyutnet.common.DEBUG_V` will print."""

SLOW_MOTION = False
"""If set, :py:attr:`soyutnet.global_defs.LOOP_DELAY` will be ``0.5``"""

label_t = int
id_t = int

if SLOW_MOTION:
    LOOP_DELAY = 0.5

TokenType = Tuple[label_t, id_t] | Tuple[Never, ...]
TokenWalletType = Dict[label_t, list[id_t]]



import asyncio
from typing import (
    Any,
    Dict,
    Tuple,
    Callable,
)

from .global_defs import *
from . import common as _c


ObserverRecordType = Tuple[float, label_t, int]
"""Observer records: (observation time, label, no of tokens with the label)"""
ObserverHistoryType = list[ObserverRecordType]


class Observer(object):
    """
    Can be assigned to a place to observe its state.
    """

    def __init__(
        self, record_limit: int = 0, verbose: bool = False, pt_ident: str = ""
    ) -> None:
        """
        Constructor.

        :param record_limit: Maximum number of records to be kept. It is unlimited if chose ``0``.
        :param verbose: Print observations if ``True``.
        :param pt_ident: Custom identification string that will be used in debug prints.
        """
        self._records: ObserverHistoryType = []
        """Records of observations. A list of triplet (observation_time, token_label, token_count)"""
        self._lock: asyncio.Lock = asyncio.Lock()
        """Async lock for safe access to records"""
        self._record_limit: int = record_limit
        """Maximum number of records to be kept."""
        self._hysteresis_bounds: Tuple[int, int] = (
            int(self._record_limit * 0.91),
            int(self._record_limit * 1.1),
        )
        """Deleting old records are optimized by an hysteresis comparator"""
        self._verbose: bool = verbose
        """Prints the observations if ``True``"""
        self._token_counters: Dict[label_t, int] = {}
        self._ident: str = pt_ident
        """Identication string that will be used in debug prints"""

    async def _clean_records(self) -> int:
        """
        Cleans records when it is required.

        :return: Number of records kept after cleaning the old ones.
        """
        count: int = -1
        if self._record_limit > 0:
            async with self._lock:
                count = len(self._records)
                if count >= self._hysteresis_bounds[1]:
                    self._records = self._records[count - self._hysteresis_bounds[0] :]
                    count = len(self._records)

        return count

    def _add_record(self, record: ObserverRecordType) -> int:
        """
        Adds a new record.

        :param record: Record.
        :return: Number of records.
        """
        self._records.append(record)

        return len(self._records)

    async def _display_records(self) -> None:
        """
        Print records to the stdout.
        """
        output: str = f"{self._ident} has {len(self._records)} records\n"
        for record in self._records:
            output += f"  {record}\n"

        _c.DEBUG(output)

    async def _save(self, record: ObserverRecordType) -> None:
        """
        Save a new record.

        :param record: Record.
        """
        if self._verbose:
            _c.DEBUG(f"REC: {self._ident}:", record)
        self._add_record(record)
        await self._clean_records()

    async def save(self, token_count_in_arc: int = 0) -> None:
        """
        Save counted tokens to the list of records.

        It is called by output transitions when they are enabled.

        :param token_count_in_arc: Number of tokens in the output arc of places must also be added to the count.
        """
        async with self._lock:
            for label in self._token_counters:
                time: float = _c.time()
                record: ObserverRecordType = (
                    time,
                    label,
                    self._token_counters[label] + token_count_in_arc,
                )
                await self._save(record)
                self._token_counters[label] = 0

    def get_records(self, column: int = -1) -> list[Any]:
        """
        Returns the records at the specified column.

        :param column: The column index of records requested. It returns all columns if it it ``-1``.
        :return: List of records.
        """
        if column < 0:
            return self._records

        output: list[Any] = []
        for rec in self._records:
            output.append(rec[column])

        return output

    async def inc_token_count(self, label: label_t, inc: int = 1) -> None:
        """
        Adds to the token count with the given label.

        It is called by places when a new token acquired or sent.

        :param label: Token label.
        :param inc: The value to be added to the count.
        """
        async with self._lock:
            if label not in self._token_counters:
                self._token_counters[label] = 0
            self._token_counters[label] += inc


class ComparativeObserver(Observer):
    """
    This observer compares the records to the provided list for test purposes.
    """

    def __init__(
        self,
        expected: Dict[int, list[Any]],
        on_comparison_ends: Callable[["ComparativeObserver"], None] | None = None,
        **kwargs: Any,
    ) -> None:
        """
        Constructor.

        :param expected: Values to be compared to.
        :param on_comparison_ends: Callback function to be called when all values in :py:attr:`self.expected` are used.
        """
        super().__init__(**kwargs)
        self._expected: Dict[int, list[Any]] = expected
        """The list of values to be compared to"""
        self._expected_index: int = 0
        """Index of the last value compared in the list"""
        self._is_still_comparing: bool = True
        """Set to ``False`` when there are no values left to compare in :py:attr:`self._expected`"""
        self._on_comparison_ends: Callable[["ComparativeObserver"], None] | None = (
            on_comparison_ends
        )
        """Callback function to be called when comparison ends"""

    async def _save(self, record: ObserverRecordType) -> None:
        """
        Save a new record after comparing.

        :param record: Record.
        :return: Number of records.
        """
        if self._is_still_comparing:
            column_count: int = 0
            for column in self._expected:
                if len(self._expected[column]) <= self._expected_index:
                    continue
                value: Any = self._expected[column][self._expected_index]
                if value >= 0 and record[column] != value:
                    raise RuntimeError(
                        f"{self._ident}: actual ({column}, {record[column]}) =/= expected ({column}, {value})"
                    )
                else:
                    _c.DEBUG_V(f"({column}, {record[column]}) == ({column}, {value})")

                if len(self._expected[column]) - 1 >= self._expected_index:
                    column_count += 1

            self._expected_index += 1
            if column_count == 0:
                self._is_still_comparing = False
                if self._verbose:
                    await self._display_records()
                if self._on_comparison_ends is not None:
                    _c.DEBUG("comparison ends")
                    self._on_comparison_ends(self)

        if self._verbose:
            _c.DEBUG(f"REC: {self._ident}:", record)
        self._add_record(record)



import asyncio
from typing import (
    Any,
    Tuple,
    Awaitable,
    Callable,
)

from .global_defs import *
from . import common as _c
from .pt_common import PTCommon, PTRegistry, Arc
from .observer import Observer


class Place(PTCommon):
    """
    Defines PTNet places.
    """

    def __init__(
        self,
        name: str = "",
        observer: Observer | None = None,
        observer_record_limit: int = 0,
        observer_verbose: bool = False,
        **kwargs: Any,
    ) -> None:
        """
        Constructor

        :param name: Name of the place.
        :param observer: Observer instance assigned to the place.
        :param observer_record_limit: Maximum number of records that is recorded by the observer. It is unlimited when chosen ``0``.
        :param observer_verbose: If set, observer will print new records when saved.
        """
        super().__init__(name=name, **kwargs)
        self._observer: Observer = (
            Observer(observer_record_limit, verbose=observer_verbose)
            if observer is None
            else observer
        )
        if not self._observer._ident:
            self._observer._ident = self.ident()

    async def _observe(self, token_count_in_arc: int = 0) -> None:
        """
        Save token counts for each label to the observer's records.

        Also, add the number of tokens sent to the caller arc.

        :param token_count_in_arc: Number of tokens in the output arc of places must also be added to the count.
        """
        await self._observer.save(token_count_in_arc)


class SpecialPlace(Place):
    """
    Custom place class whose token processing methods can be overriden.
    """

    def __init__(
        self,
        name: str = "",
        consumer: Callable[["SpecialPlace"], Awaitable[bool]] | None = None,
        producer: Callable[["SpecialPlace"], Awaitable[TokenType]] | None = None,
        **kwargs: Any,
    ) -> None:
        """
        Constructor.

        :param name: Name of the place.
        :param consumer: Custom :py:func:`place._process_input_arcs` function.
        :param producer: Custom :py:func:`place._process_output_arcs` function.
        """
        super().__init__(name=name, **kwargs)
        self._consumer: Callable[["SpecialPlace"], Awaitable[bool]] | None = consumer
        """Custom :py:func:`place._process_input_arcs` function."""
        self._producer: Callable[["SpecialPlace"], Awaitable[TokenType]] | None = (
            producer
        )
        """Custom :py:func:`place._process_output_arcs` function."""

    async def _process_input_arcs(self) -> bool:
        """
        Calls custom producer function. If it is ``None`` calls :py:func:`place._process_input_arcs` function.

        :return: If ``True`` continues to processing tokens and output arc, else loops back to processing input arcs.
        """
        if self._producer is not None:
            token: TokenType = await self._producer(self)
            if not token:
                return False

            label: label_t = token[0]
            count: int = self._put_token(token, strict=False)
            if self._observer is not None:
                await self._observer.inc_token_count(token[0])

            return True
        else:
            return await super()._process_input_arcs()

    async def _process_output_arcs(self) -> None:
        """
        Null process output arcs function, when a custom consumer is defined.
        """
        if self._consumer is not None:
            await self._consumer(self)
            return

        await super()._process_output_arcs()

    async def _process_tokens(self) -> bool:
        """
        Calls custom consumer function. If it is ``None`` calls :py:func:`place._process_tokens` function.
        """
        if not await super()._process_tokens():
            return False

        if self._observer is not None:
            """Emulate how it will behave when tokens are sent over an input arc with weight=1."""
            token_count_in_arc = 0
            if self._producer is not None:
                token_count_in_arc = 1
            await self._observe(token_count_in_arc=token_count_in_arc)

        return True

    async def observe(self, token_count_in_arc: int = 0) -> None:
        return



import sys
import asyncio

import soyutnet
from soyutnet.pt_common import PTRegistry
from soyutnet.token import Token
from soyutnet.place import SpecialPlace, Place
from soyutnet.transition import Transition
from soyutnet.common import GENERIC_ID, GENERIC_LABEL, DEBUG
from soyutnet.observer import ComparativeObserver, Observer


def co_begin(action_count):
    token_ids = list(range(1, action_count + 1))

    async def producer(place):
        try:
            id: id_t = token_ids.pop(0)
            return (GENERIC_LABEL, id)
        except IndexError:
            pass

        return tuple()

    output_token_ids = list(token_ids)

    async def consumer(place):
        token = place.get_token(GENERIC_LABEL)
        if not token:
            return
        output_token_ids.remove(token[1])
        DEBUG(f"Token '{token}' is removed.")
        if not output_token_ids:
            soyutnet.terminate()

    reg = PTRegistry()
    p0 = SpecialPlace("p0", producer=producer)
    t0 = Transition("t0")
    p0.connect(t0, weight=action_count)
    reg.register(p0)
    reg.register(t0)
    for i in range(action_count):
        p1i = SpecialPlace(f"p1_{i}", consumer=consumer)
        reg.register(p1i)
        t0.connect(p1i)

    asyncio.run(soyutnet.main(reg))


def co_end(action_count):
    async def producer(place):
        DEBUG(f"{place.ident()}")
        place_id = place.ident().split(",")[0].split("_")[1]
        try:
            return (GENERIC_LABEL, int(place_id))
        except IndexError:
            pass

        return tuple()

    output_token_ids = list(range(1, action_count + 1))

    async def consumer(place):
        token = place.get_token(GENERIC_LABEL)
        if not token:
            return
        output_token_ids.remove(token[1])
        DEBUG(f"Token '{token}' is removed.")
        if not output_token_ids:
            soyutnet.terminate()

    reg = PTRegistry()
    p1 = SpecialPlace("p0", consumer=consumer)
    t0 = Transition("t0")
    t0.connect(p1)
    reg.register(p1)
    reg.register(t0)
    for i in range(action_count):
        p0i = SpecialPlace(f"p0_{i+1}", producer=producer)
        reg.register(p0i)
        p0i.connect(t0)

    asyncio.run(soyutnet.main(reg))


def sync_by_signal():
    token_ids = list(range(1, 11)) + [GENERIC_ID] * 20

    async def producer(place):
        try:
            id = token_ids.pop(0)
            DEBUG(f"Produced '{(GENERIC_LABEL, id)}'")
            return (GENERIC_LABEL, id)
        except IndexError:
            pass

        return tuple()

    output_token_ids = token_ids[:10]

    async def consumer(place):
        token = place.get_token(GENERIC_LABEL)
        if not token:
            return
        if token[1] == GENERIC_ID:
            return
        output_token_ids.remove(token[1])
        DEBUG(f"Consumed '{token}'")
        if not output_token_ids:
            soyutnet.terminate()

    activate = False

    async def processor(place):
        nonlocal activate
        if place._get_token_count(GENERIC_LABEL) > 3:
            DEBUG(f"Activated. 'p4' will start consuming.")
            activate = True

        return activate

    p0 = SpecialPlace("p0", producer=producer)
    p1 = SpecialPlace("p1", consumer=consumer)
    p2 = Place("p2", processor=processor)
    p3 = SpecialPlace("p3", producer=producer)
    p4 = SpecialPlace("p4", consumer=consumer)
    t0 = Transition("t0")
    t1 = Transition("t1")

    reg = PTRegistry()
    reg.register(p0)
    reg.register(p1)
    reg.register(p2)
    reg.register(p3)
    reg.register(p4)
    reg.register(t0)
    reg.register(t1)

    p0.connect(t0, weight=2).connect(p1)
    t0.connect(p2).connect(t1)
    p3.connect(t1).connect(p4)

    asyncio.run(soyutnet.main(reg))


def feedback(N=1):
    token_ids = list(range(1, 5))
    WRITER_LABEL = 1

    async def producer(place):
        try:
            id = token_ids.pop(0)
            label = WRITER_LABEL
        except IndexError:
            id = GENERIC_ID
            label = GENERIC_LABEL
        token = (label, id)
        DEBUG(f"Produced {token}")
        return token

    consumed_ids = list(token_ids)

    async def consumer(place):
        nonlocal consumed_ids
        token = place.get_token(WRITER_LABEL)
        if not token:
            return
        try:
            consumed_ids.remove(token[1])
        except ValueError:
            return
        print(f"Consumed {token}")
        if not consumed_ids:
            soyutnet.terminate()

    p0 = SpecialPlace("p0", producer=producer)
    pt0 = Transition("pt0")
    p1 = Place("p1", observer_verbose=False)
    pt1 = Transition("pt1")
    p2 = Place("p2", observer_verbose=False)
    pt2 = Transition("pt2")
    p3 = SpecialPlace("p3", consumer=consumer)

    FEEDBACK_LABEL = 1
    initial_tokens = {
        GENERIC_LABEL: [GENERIC_ID] * (1 + N),
    }
    lock = Place("lock", initial_tokens=initial_tokens)

    reg = PTRegistry()
    reg.register(p0)
    reg.register(p1)
    reg.register(p2)
    reg.register(p3)
    reg.register(pt0)
    reg.register(pt1)
    reg.register(pt2)
    reg.register(lock)

    labels = [WRITER_LABEL, GENERIC_LABEL]
    (
        p0.connect(pt0, labels=labels)
        .connect(p1, labels=labels)
        .connect(pt1, labels=labels)
        .connect(p2, weight=2, labels=labels)
        .connect(pt2, weight=N + 1, labels=labels)
        .connect(p3, labels=labels[:1])
    )

    pt2.connect(lock, weight=N).connect(pt1)

    try:
        asyncio.run(soyutnet.main(reg))
    except asyncio.exceptions.CancelledError:
        pass

    print(pt0.ident(), pt0.get_no_of_times_enabled())
    print(pt1.ident(), pt1.get_no_of_times_enabled())
    print(pt2.ident(), pt2.get_no_of_times_enabled())


def reader_writer(N=1):
    writer_produced_count = 0
    WRITER_LABEL = 1

    async def writer_producer(place):
        nonlocal writer_produced_count
        writer_produced_count += 1
        DEBUG(f"Produced {writer_produced_count}")
        return (WRITER_LABEL, writer_produced_count)

    reader_produced_count = 0
    READER_LABEL = 2

    async def reader_producer(place):
        nonlocal reader_produced_count
        reader_produced_count += 1
        DEBUG(f"Produced {reader_produced_count}")
        return (READER_LABEL, reader_produced_count)

    consumed_count = {WRITER_LABEL: 5, READER_LABEL: 5}

    async def consumer(place):
        nonlocal consumed_count
        token = place.get_token(WRITER_LABEL) or place.get_token(READER_LABEL)
        if not token:
            return
        consumed_count[token[0]] -= 1
        DEBUG(f"Consumed {consumed_count}, {token}")
        if sum(consumed_count.values()) == 0:
            soyutnet.terminate()

    w0 = SpecialPlace("w0", producer=writer_producer)
    wt0 = Transition("wt0")
    w1 = Place("w1", observer_verbose=True)
    wt1 = Transition("wt1")
    w2 = Place("w2", observer_verbose=True)
    wt2 = Transition("wt2")
    w3 = SpecialPlace("w3", consumer=consumer)

    r0 = SpecialPlace("r0", producer=reader_producer)
    rt0 = Transition("rt0")
    r1 = Place("r1", observer_verbose=True)
    rt1 = Transition("rt1")
    r2 = Place("r2", observer_verbose=True)
    rt2 = Transition("rt2")
    r3 = SpecialPlace("r3", consumer=consumer)

    initial_tokens = {
        WRITER_LABEL: [GENERIC_ID] * (1),
        READER_LABEL: [GENERIC_ID] * (1),
    }
    lock = Place("lock", initial_tokens=initial_tokens)

    reg = PTRegistry()
    reg.register(w0)
    reg.register(w1)
    reg.register(w2)
    reg.register(w3)
    reg.register(wt0)
    reg.register(wt1)
    reg.register(wt2)
    # reg.register(r0)
    # reg.register(r1)
    # reg.register(r2)
    # reg.register(r3)
    # reg.register(rt0)
    # reg.register(rt1)
    # reg.register(rt2)
    reg.register(lock)

    (
        w0.connect(wt0, label=WRITER_LABEL)
        .connect(w1, label=WRITER_LABEL)
        .connect(wt1, label=WRITER_LABEL)
        .connect(w2, label=WRITER_LABEL, weight=2)
        .connect(wt2, label=WRITER_LABEL, weight=2)
        .connect(w3, label=WRITER_LABEL)
    )

    # (r0.connect(rt0, label=READER_LABEL)
    # .connect(r1, label=READER_LABEL)
    # .connect(rt1, label=READER_LABEL)
    # .connect(r2, label=READER_LABEL)
    # .connect(rt2, label=READER_LABEL, weight=1)
    # .connect(r3, label=READER_LABEL))

    wt2.connect(lock, label=WRITER_LABEL, weight=N).connect(
        wt1, label=WRITER_LABEL, weight=N
    )
    # rt2.connect(lock, label=READER_LABEL).connect(rt1, label=READER_LABEL)

    asyncio.run(soyutnet.main(reg))


if __name__ == "__main__":
    # co_end(int(sys.argv[1]))
    sync_by_signal()
    # feedback(int(sys.argv[1]))



import asyncio

from typing import (
    Any,
)

from .global_defs import *
from . import common as _c
from .registry import (
    Registry,
)


class Token(object):
    def __init__(self, label: label_t = _c.GENERIC_LABEL, binding: Any = None) -> None:
        """
        A class that represents a labeled PTNet token which can be binded to any object.

        :param label: Label.
        :param binding: Binded object.
        """
        self._label: label_t = label
        """Token label"""
        self._id: id_t = _c.GENERIC_ID
        """Unique token ID"""
        self._binding: Any = binding
        """Binded object"""
        self._ident: str = ""  # TODO: Place ident logic here remove from other parts

    def get_label(self) -> label_t:
        """
        Returns the label of token

        :return: Label.
        """
        return self._label

    def get_binding(self) -> Any:
        """
        Returns the object binded to the token.

        :return: Binded object.
        """
        return self._binding

    def get_id(self) -> id_t:
        """
        Returns token's unique ID.

        :return: ID.
        """
        return self._id


class TokenRegistry(Registry):
    def __init__(self) -> None:
        super().__init__()

    def register(self, token: Token) -> id_t:  # type: ignore[override]
        """
        Register a new token

        :param token: New token.
        :return: Assigned unique ID.
        """

        def callback(new_id: id_t, tkn: Any) -> None:

            tkn._id = new_id

        return super().register(token, callback)



import asyncio
from typing import (
    Any,
    Dict,
    Callable,
    Awaitable,
    Generator,
    Tuple,
)

from .global_defs import *
from . import common as _c

DirectoryType = Dict[label_t, list[Tuple[id_t, Any]]]
PostRegisterCallbackType = Callable[[id_t, Any], None]


def _default_post_register_callback(dummy1: Any, dummy2: int) -> None:
    pass


class Registry(object):
    """
    Registry keeps track of (label, id) tuples and the objects assigned to them.
    It generates unique ids for new objects.
    """

    def __init__(self) -> None:
        self._id_counter: id_t = _c.INITIAL_ID
        """Auto-incrementing id assigned to new objects"""
        self._directory: DirectoryType = {}
        """Keeps all objects categorized by labels"""
        self._lock: asyncio.Lock = asyncio.Lock()
        """Locks access to :py:attr:`self._directory`"""

    def _new_id(self) -> id_t:
        """
        Creates new ids for new objects.

        :return: Unique id
        """
        self._id_counter += 1
        return self._id_counter

    def register(
        self,
        obj: Any,
        post_register_callback: PostRegisterCallbackType = _default_post_register_callback,
    ) -> id_t:
        """
        Register a new object

        :param obj: New object of any type.
        :param post_register_callback: Called after object is registered.
        :return: Assigned unique ID.
        """
        new_id: id_t = self._new_id()
        label: label_t = obj.get_label()
        if label not in self._directory:
            self._directory[label] = []
        self._directory[label].append((new_id, obj))
        if post_register_callback != _default_post_register_callback:
            post_register_callback(new_id, obj)

        return new_id

    def get_entry_count(self, label: label_t = _c.GENERIC_LABEL) -> int:
        """
        Returns the number of entries with the given label.

        :param label: Label.
        :return: Number of entries.
        """
        if label in self._directory:
            return len(self._directory)

        return 0

    def get_first_entry(self, label: label_t = _c.GENERIC_LABEL) -> Tuple[id_t, Any]:
        """
        Returns first entry with given label. First entry is the one registered first.

        :param label: Label.
        :return: Entry.
        """
        if label in self._directory and len(self._directory[label]) > 0:
            return self._directory[label][0]

        return (_c.INVALID_ID, None)

    def get_entries(self, label: label_t, id: id_t) -> list[Any]:
        """
        Returns a list of objects with given label and ID.

        :param label: Label.
        :param id: ID.
        :return: A list of objects.
        """
        result: list[Any] = []
        if label not in self._directory:
            return result

        for entry in self._directory[label]:
            if entry[0] == id:
                result.append(entry[1])

        return result

    def entries(self, label: label_t | None = None) -> Generator[Any, None, None]:
        """
        Iterates through all entries with the given label.

        :param label: Label. If it is ``None``, iterates through all labels.
        :return: Yields entries
        """
        d: DirectoryType = (
            {label: self._directory[label]}
            if label in self._directory
            else self._directory
        )
        for label in d:
            for entry in d[label]:
                yield entry



import asyncio

import soyutnet
from soyutnet.pt_common import PTRegistry
from soyutnet.token import Token
from soyutnet.place import SpecialPlace, Place
from soyutnet.transition import Transition
from soyutnet.common import GENERIC_ID, GENERIC_LABEL
from soyutnet.observer import ComparativeObserver, Observer


def main():
    token_ids = list(range(1, 7))

    async def producer(place):
        try:
            id: id_t = token_ids.pop(0)
            token = (GENERIC_LABEL, id)
            print("Produced:", token)
            return token
        except IndexError:
            pass

        return tuple()

    async def consumer(place):
        token = place.get_token(GENERIC_LABEL)
        if token:
            print("Consumed:", token)
        else:
            print("No token in consumer")

    place_count = 2

    def on_comparison_ends(observer):
        nonlocal place_count
        place_count -= 1
        if place_count == 0:
            soyutnet.terminate()

    o1 = ComparativeObserver(
        expected={2: [-1] + [1] * 4},
        on_comparison_ends=on_comparison_ends,
        verbose=False,
    )
    o2 = ComparativeObserver(
        expected={2: [1] * 5},
        on_comparison_ends=on_comparison_ends,
        verbose=False,
    )
    p1 = SpecialPlace("p1", producer=producer, observer=o1)
    p2 = SpecialPlace("p2", consumer=consumer, observer=o2)
    t1 = Transition("t1")

    reg = PTRegistry()

    reg.register(p1)
    reg.register(p2)
    reg.register(t1)

    p1.connect(t1).connect(p2)

    try:
        asyncio.run(soyutnet.main(reg))
    except asyncio.exceptions.CancelledError:
        print("Simulation is terminated.")


if __name__ == "__main__":
    main()



import asyncio
from typing import Any

from .global_defs import *


INVALID_LABEL = -10
INVALID_ID = -11
GENERIC_LABEL = 0
"""Generic label"""
GENERIC_ID = 0
"""Generic ID"""
INITIAL_ID = 0


class SoyutNetError(Exception):
    def __init__(self, message: str = "An error occured.") -> None:
        self.message: str = message
        super().__init__(self.message)


async def sleep(amount: float = 0.0) -> None:
    await asyncio.sleep(amount)


def time() -> float:
    loop = asyncio.get_running_loop()
    return loop.time()


def get_loop_name() -> str:
    name: str = "NO-LOOP"
    try:
        task: asyncio.Task[Any] | None = asyncio.current_task()
        if isinstance(task, asyncio.Task):
            name = task.get_name()
    except RuntimeError:
        pass

    return name


def _DEBUG(*args: Any) -> None:
    pass


def _DEBUG_V(*args: Any) -> None:
    print(f"{get_loop_name()}:", *args)


def ERROR_V(*args: Any) -> None:
    _DEBUG_V("ERR:", *args)


def _DEBUG_NOP(*args: Any) -> None:
    pass


if DEBUG_ENABLED:
    DEBUG = _DEBUG_V
    VERBOSE_DEBUG_ENABLED = VERBOSE_ENABLED
else:
    DEBUG = _DEBUG_NOP
    VERBOSE_DEBUG_ENABLED = False

if VERBOSE_DEBUG_ENABLED:
    DEBUG_V = _DEBUG_V
else:
    DEBUG_V = _DEBUG_NOP

if not VERBOSE_ENABLED:
    ERROR_V = _DEBUG_NOP



import pytest
import asyncio

import soyutnet
from soyutnet.token import (
    Token,
    TokenRegistry,
)
from soyutnet.pt_common import PTRegistry
from soyutnet.place import (
    Place,
    SpecialPlace,
)
from soyutnet.transition import (
    Transition,
)
from soyutnet.common import (
    GENERIC_LABEL,
    GENERIC_ID,
    INITIAL_ID,
    INVALID_ID,
)


@pytest.mark.asyncio
async def test_01():
    registry = TokenRegistry()
    token = Token()

    assert (INVALID_ID, None) == registry.get_first_entry(token.get_label())
    registry.register(token)
    assert token.get_id() == INITIAL_ID + 1
    assert token.get_label() == GENERIC_LABEL
    assert (token.get_id(), token) == registry.get_first_entry(token.get_label())

    assert registry.get_entry_count() == 1


@pytest.mark.asyncio
async def test_02():
    place = Place()

    assert place.get_id() == GENERIC_ID
    assert place.get_binding() is None


@pytest.mark.asyncio
async def test_03():
    transition = Transition()

    assert transition.get_id() == GENERIC_ID
    assert transition.get_binding() is None


def test_04():
    import simple_example as e

    e.main()


def test_05():
    import simple_example_different_weight as e

    for i in range(1, 15):
        for j in range(1, i + 1):
            try:
                e.main(i, j)
            except asyncio.exceptions.CancelledError:
                pass


def test_06():
    import simple_example_two_input_places as e

    for i in range(100, 10000, 1000):
        try:
            e.main(i)
        except asyncio.exceptions.CancelledError:
            pass


def test_07():
    import simple_example_two_input_places_but_different_weights as e

    MAX = 10
    for i in range(1, MAX + 1):
        for j in range(1, MAX + 1):
            for k in range(1, MAX + 1):
                e.main(w1=i, w2=j)


def test_08():
    from basic_models import co_begin

    for i in range(2, 100):
        try:
            co_begin(i)
        except asyncio.exceptions.CancelledError:
            pass


def test_09():
    from basic_models import co_end

    for i in range(2, 100):
        try:
            co_end(i)
        except asyncio.exceptions.CancelledError:
            pass


def test_10():
    from basic_models import sync_by_signal

    try:
        sync_by_signal()
    except asyncio.exceptions.CancelledError:
        pass



import sys
import asyncio

import soyutnet
from soyutnet.pt_common import PTRegistry
from soyutnet.token import Token
from soyutnet.place import SpecialPlace, Place
from soyutnet.transition import Transition
from soyutnet.common import GENERIC_ID, GENERIC_LABEL
from soyutnet.observer import ComparativeObserver, Observer
from soyutnet.common import DEBUG


def main(w1=1, w2=1):
    token_ids = [GENERIC_ID] * (w1 * w2 * 10)

    async def producer(place):
        try:
            id: id_t = token_ids.pop(0)
            return (GENERIC_LABEL, id)
        except IndexError:
            pass

        return tuple()

    async def consumer(place):
        return

    place_count = 3

    def on_comparison_ends(observer):
        nonlocal place_count
        place_count -= 1
        if place_count == 0:
            soyutnet.terminate()

    expected1 = {2: [w1] * (len(token_ids) // w1 - 1)}
    expected2 = {2: [w2 * int(i % w1 == 0) for i in range(len(token_ids) - (w1 * w2))]}

    o0 = ComparativeObserver(
        expected={2: [-1] + [1] * (len(token_ids) - 2)},
        on_comparison_ends=on_comparison_ends,
        verbose=False,
    )
    o1 = ComparativeObserver(
        expected=expected1,
        on_comparison_ends=on_comparison_ends,
        verbose=False,
    )
    o2 = ComparativeObserver(
        expected=expected2,
        on_comparison_ends=on_comparison_ends,
        verbose=True,
    )
    p0 = SpecialPlace("p0", producer=producer, observer=o0)
    p1 = Place("p1", observer=o1)
    p2 = SpecialPlace("p2", consumer=consumer, observer=o2)
    t1 = Transition("t1")
    t2 = Transition("t2")

    reg = PTRegistry()

    reg.register(p0)
    reg.register(p1)
    reg.register(p2)
    reg.register(t1)
    reg.register(t2)

    p0.connect(t1)
    t1.connect(p1)
    p1.connect(t2, weight=w1)
    t2.connect(p2, weight=w2)

    asyncio.run(soyutnet.main(reg))


if __name__ == "__main__":
    main(int(sys.argv[1]), int(sys.argv[2]))
